import java.util.*;
import java.io.*;

/**
 * Just a small tool to convert files from CSV to LDIF format.
 * @version 1.0
 * 
 * This program is free software, covered by the GPL license (get your
 * copy from: <a href="http://www.gnu.org">www.gnu.org</a>). 
 * It was written by 
 * <a href="mailto:p.ahlbrecht@billiton.de">Patrick Ahlbrecht</a>
 *
 * Format of the CSV file:
 * - Empty lines or lines starting with CSIGN are ignored
 * - Every value must be enclosed in ENC
 * - Values must be seperated by DELIM
 * - Its ok to have DELIM's following a DELIM, these are regarded as empty
 *   Values
 * - The first CSV line in the file will be used for the left handside of
 *   the ldif entries
 * - A line starting with CONF will be used for configuration
 * - A value may include newline characters (though not recommended)
 * 
 * Example:
 * 
 * # File starts here
 * @uid@ou=people, o=myorg@top@person
 * "uid","sn","cn"
 * "joe","User","Joe User"
 *  # File ends here
 *
 * Will lead to the following ldif:
 *
 * dn: uid=joe,ou=people, o=myorg
 * objectClass: top
 * objectClass: person
 * uid: joe
 * sn: User
 * cn: Joe User
 */
public class csv2ldif {
  
  /**
   * Delimeter constant. commas, tabs ? Doesn't matter
   */
  public static final char DELIM=',';
  
  /**
   * What values are enclosed in. Quotes are reasonable
   */
  public static final char ENC='\"';
  
  /**
   * How to mark lines as comments
   */
  public static final String CSIGN="#";
  
  /**
   * How to mark config lines
   */
  public static final String CONF="@";
  
  /**
   * What should a linebreak in a multilined record be tranlated into ?
   */
  public static final String LINEBREAK="\n ";
  
  /**
   * The "path part" of your distinguished name
   */
  public static String dn_path;
  
  /**
   * The "unique id part" of your distinguished name
   */
  public static String dn_id;
  
  /**
   * Buffer for assembling an ldif entry
   */
  public static StringBuffer entry;
  
  /**
   * Buffer for ldif preamble (e.g.: objectClass:)
   */
  public static String preamble="";
  
  /**
   * Print Usage information
   */
  public static void printUsage() {
    System.out.println("Convert files from csv to ldif");
    System.out.println("");
    System.out.println("Usage: Just give the csv files as arguments");
  }
  
  /**
   * Write an ldif entry to stdout (params will be cleaned of non-conform
   * characters).
   * @param d lefthandside of the entry
   * @param v righthandside of the entry
   */
  public static void printEntry(String d,String v) {
    if ((d==null)||(v==null)) return;
    entry.append(d);
    entry.append(": ");
    entry.append(v);
    entry.append("\n");
    if (d.equals(dn_id)) {
      String tmp="dn: "+dn_id+"="+v+", "+dn_path+"\n"+preamble+entry.toString();
      entry=new StringBuffer(tmp);
    }
  }
  
  /**
   * parse CSV strings, grep values
   * @param line a line of text in CSV format
   * @return array of strings, containing the values
   */
  public static String[] getValues(String line) {
    int elems=0;
    char[] c=line.toCharArray();
    if (!(c[0]==ENC) || (c[0]==DELIM)) {
      return null;
    }
    
    for(int i=0;i<c.length;i++) {
      if (c[i]==DELIM) {
        elems++;
      }
      else {
       int a;
       for (a=1;a<c.length-i;a++) {
         if (c[a+i]==ENC) break;
       }
       i=i+a+1;
       elems++;
      }
    } 

    String[] val = new String[elems];
    elems=0;
    for(int i=0;i<c.length;i++) {
      if (c[i]==DELIM) {
        val[elems]=null;
        elems++;
      }
      else {
       int a;
       for (a=1;a<c.length-i;a++) {
         if (c[a+i]==ENC) break;
       }
       char[] tmp= new char[a+1];
       System.arraycopy(c,i,tmp,0,a+1);
       String s= new String(tmp).substring(1);
       s=s.substring(0,s.length()-1);
       if (s.equals("")) val[elems]=null; 
       else val[elems]=s;
       i=i+a+1;
       elems++;
      }
    }
    return val;
  }
  
  /**
   * Tests for multi lined records
   * @param l the line to test
   * @return true at the line contains a multilined value
   */
  public static boolean multiRecord(String l) {
    char[] c= l.toCharArray();
    int count=0;
    for (int i=0;i<c.length;i++) {
      if (c[i]==ENC) count++;
    }
    if ((count%2)>0) return true;
    return false;
  }
  
  /**
   * Converts a file from CSV to LDIF
   * @param fname the filename, containing the CSV data, LDIF will be written
   * to stdout
   */
  public static void convert(String fname) {
    int lc=0;
    try {
      BufferedReader in=new BufferedReader(new FileReader(fname));
      String[] desc=null;
      String line;
      System.out.println("### Entry from file: "+fname);
      while((line=in.readLine())!=null) {
        lc++;
        entry=new StringBuffer();
        if (line.startsWith(CONF)) {
          preamble="";
          StringTokenizer st = new StringTokenizer(line,CONF);
          dn_id=st.nextToken();
          dn_path=st.nextToken();
          while (st.hasMoreTokens()) {
            preamble=preamble+"objectClass: "+st.nextToken()+"\n";
          }
          line="";
        }
        if (!((line.equals("")) || (line.startsWith(CSIGN)))) {
          // Multiline gives you a headache, but I've got Aspirine ;-).
          while (multiRecord(line)) {
            
            line=line+LINEBREAK+in.readLine();
          }
          if (desc==null) {
            desc=getValues(line);
            //if (dn_id==null) dn_id=desc[0];
          }
          else {
            String[] val=getValues(line);
            if (val!=null) {
              for(int i=0;i<val.length;i++) { 
                /*
                // Ignore surplus fields
                if (i>desc.length-1) {
                  System.err.println("### Line: "+lc+" v/d missmatch");
                  break;
                }*/
                try {
                  printEntry(desc[i],val[i]);
                }
                catch(ArrayIndexOutOfBoundsException e) {
                  printEntry("# Field"+i,val[i]);
                }
              }
              //entry.append("### End of record ("+lc+")\n\n");
              System.out.println(entry.toString());
            }
          }
        }
      }
      in.close();
    }
    catch (IOException e) {
      System.err.println("### Error reading file: "+fname);
    }
    catch (Exception e) {
      System.err.println("### Program Error (line "+lc+"): ");
      e.printStackTrace(System.err);
    }
  }
  
  
  /**
   * Just guess ;-)
   */
  public static void main(String[] args) {
    if ((args.length==0) || args[0].equals("--help")) {
      printUsage();
      System.exit(1);
    }
    for(int i=0;i<args.length;i++) {
      convert(args[i]);
    }
  }
}